#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import tempfile

from python_base_app import audio_handler
from python_base_app import configuration
from python_base_app import log_handling
from python_base_app.test import base_test

SPOOL_DIR = tempfile.gettempdir()
TEXT = "hallo"
LOCALE = "de_DE"


class TestAudioHandler(base_test.BaseTestCase):

    @staticmethod
    def delete_audio_file(p_audio_handler):

        audio_file = p_audio_handler.get_audio_filename(p_text=TEXT, p_locale=None)

        try:
            os.unlink(audio_file)

        except Exception as e:
            logger = log_handling.get_logger()
            fmt = "Cannot delete audio file {filename}: {exception}"
            logger.warning(fmt.format(filename=audio_file, exception=str(e)))

    @base_test.skip_if_env("NO_AUDIO_OUTPUT")
    def test_engine_google_init(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = audio_handler.SPEECH_ENGINE_GOOGLE
        a_handler = audio_handler.AudioHandler(p_config=a_config)

        self.assertIsNotNone(a_handler)

    @base_test.skip_if_env("NO_AUDIO_OUTPUT")
    def test_engine_google_speak(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = audio_handler.SPEECH_ENGINE_GOOGLE
        a_config.spool_dir = SPOOL_DIR
        a_config.locale = LOCALE
        a_handler = audio_handler.AudioHandler(p_config=a_config)

        self.assertIsNotNone(a_handler)

        self.delete_audio_file(p_audio_handler=a_handler)
        a_handler.notify(p_text=TEXT, p_locale=LOCALE)
        a_handler.notify(p_text=TEXT, p_locale=LOCALE)
        a_handler.stop_engine()

    @base_test.skip_if_env("NO_AUDIO_OUTPUT")
    def test_spool_dir_and_file(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = audio_handler.SPEECH_ENGINE_GOOGLE
        a_config.spool_dir = SPOOL_DIR
        a_handler = audio_handler.AudioHandler(p_config=a_config)

        self.assertIsNotNone(a_handler)

        audio_file = a_handler.get_audio_filename(p_text=TEXT, p_locale=None)

        self.assertTrue(audio_file.startswith(SPOOL_DIR))

        self.delete_audio_file(p_audio_handler=a_handler)
        a_thread = a_handler.notify(p_text=TEXT)
        a_thread.join()
        #        time.sleep(1)

        self.assertTrue(os.path.exists(audio_file))

        a_handler.stop_engine()

    def test_engine_external_init(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = audio_handler.SPEECH_ENGINE_EXTERNAL
        a_handler = audio_handler.AudioHandler(p_config=a_config)

        self.assertIsNotNone(a_handler)

        a_handler.stop_engine()

    @base_test.skip_if_env("NO_AUDIO_OUTPUT")
    def test_engine_external_speak(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = audio_handler.SPEECH_ENGINE_EXTERNAL
        a_config.spool_dir = SPOOL_DIR
        a_handler = audio_handler.AudioHandler(p_config=a_config)

        self.assertIsNotNone(a_handler)

        self.delete_audio_file(p_audio_handler=a_handler)
        a_handler.notify(p_text=TEXT)

        a_handler.stop_engine()

    @base_test.skip_if_env("NO_AUDIO_OUTPUT")
    def test_engine_external_speak_mixer(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = audio_handler.SPEECH_ENGINE_EXTERNAL
        a_config.spool_dir = SPOOL_DIR
        a_config.audio_mixer_volume = 50
        a_handler = audio_handler.AudioHandler(p_config=a_config)

        self.assertIsNotNone(a_handler)

        self.delete_audio_file(p_audio_handler=a_handler)
        a_handler.notify(p_text=TEXT)

        a_handler.stop_engine()

    def test_engine_invalid(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_config.speech_engine = "NOT-EXISTING"

        with self.assertRaises(configuration.ConfigurationException):
            a_handler = audio_handler.AudioHandler(p_config=a_config)
            self.assertIsNotNone(a_handler)




#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import sys
import tempfile

import setuptools

from python_base_app import base_ci_toolbox
from python_base_app.test import base_test

SPOOL_DIR = tempfile.gettempdir()
TEXT = "hallo"
LOCALE = "de_DE"

STAGE_PREPARE_FILES = [
    '.circleci/config.yml',
    '.gitlab-ci.yml',
    '.codacy.yml',
    'a_package/git.py',
    'bin/pip3.sh'
]

STAGE_PREPARE_DIRS = [
    '.circleci',
    'bin'
]

STAGE_BUILD_FILES = [
    'a_package/translations/de/LC_MESSAGES/messages.mo',
    'a_package.egg-info/dependency_links.txt',
    'a_package.egg-info/PKG-INFO',
    'a_package.egg-info/requires.txt',
    'a_package.egg-info/SOURCES.txt',
    'a_package.egg-info/top_level.txt',
    'dist/a_package-0.0.1.tar.gz',
    'bin/make-debian-package.sh',
    'bin/pip3.sh'
]

STAGE_BUILD_DIRS = [
    'bin',
    'dist',
    'a_package.egg-info'
]

STAGE_TEST_FILES = [
    'bin/test-app.sh',
    '.coveragerc',
]

STAGE_TEST_DIRS = [
    'bin'
]

RESOURCE_BASE_PATH = "resources/ci_toolbox"


class TestBaseCiToolbox(base_test.BaseTestCase):

    def get_resource_base_path(self):

        return os.path.join(os.path.dirname(__file__), RESOURCE_BASE_PATH)

    def remove_files(self, p_rel_paths, p_rel_dirs):

        for rel_path in p_rel_paths:
            self.rel_delete_file(rel_path)

        for rel_dir in p_rel_dirs:
            self.rel_delete_dir(rel_dir)

    def rel_test_files(self, p_rel_paths):

        for rel_path in p_rel_paths:
            dir_path = os.path.join(self.get_resource_base_path(), rel_path)

            self.assertTrue(os.path.exists(dir_path), "file {path} exists".format(path=dir_path))

    def rel_delete_dir(self, p_rel_path):
        dir_path = os.path.join(self.get_resource_base_path(), p_rel_path)

        if os.path.exists(dir_path):
            os.rmdir(dir_path)

    def rel_delete_file(self, p_rel_path):
        dir_path = os.path.join(self.get_resource_base_path(), p_rel_path)

        if os.path.exists(dir_path):
            os.unlink(dir_path)

    def test_stage_prepare(self):

        self.remove_files(p_rel_paths=STAGE_PREPARE_FILES, p_rel_dirs=STAGE_PREPARE_DIRS)
        main_module_dir = os.path.join(os.path.dirname(__file__), self.get_resource_base_path())

        sys.argv.extend(['--execute-stage', 'PREPARE'])
        base_ci_toolbox.main(p_main_module_dir=main_module_dir)

        self.rel_test_files(p_rel_paths=STAGE_PREPARE_FILES)

        self.remove_files(p_rel_paths=STAGE_PREPARE_FILES, p_rel_dirs=STAGE_PREPARE_DIRS)

    def test_stage_build(self):

        print("setuptools.version", setuptools.version.__version__)

        self.remove_files(p_rel_paths=STAGE_BUILD_FILES, p_rel_dirs=STAGE_BUILD_DIRS)
        main_module_dir = os.path.join(os.path.dirname(__file__), self.get_resource_base_path())

        sys.argv.extend(['--execute-stage', 'BUILD', '--use-dev-dir', self.get_resource_base_path()])
        base_ci_toolbox.main(p_main_module_dir=main_module_dir)

        self.rel_test_files(p_rel_paths=STAGE_BUILD_FILES)

        self.remove_files(p_rel_paths=STAGE_BUILD_FILES, p_rel_dirs=STAGE_BUILD_DIRS)



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import unittest

from python_base_app import log_handling
from python_base_app.test import base_test
from python_base_app.test import test_audio_handler
from python_base_app.test import test_audio_player
from python_base_app.test import test_base_ci_toolbox
from python_base_app.test import test_base_web_server
from python_base_app.test import test_configuration
from python_base_app.test import test_pytest
from python_base_app.test import test_unix_user_handler


def add_test_cases(p_test_suite, p_config_filename=None):
    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_configuration.TestConfiguration, p_config_filename=p_config_filename)

    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_audio_handler.TestAudioHandler, p_config_filename=p_config_filename)

    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_audio_player.TestAudioPlayer, p_config_filename=p_config_filename)

    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_pytest.TestPytest, p_config_filename=p_config_filename)

    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_base_ci_toolbox.TestBaseCiToolbox, p_config_filename=p_config_filename)

    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_base_web_server.TestBaseWebServer, p_config_filename=p_config_filename)

    base_test.add_tests_in_test_unit(
        p_test_suite=p_test_suite,
        p_test_unit_class=test_unix_user_handler.TestUnixUserHandler, p_config_filename=p_config_filename)


def main():
    log_handling.start_logging(p_use_filter=False)
    test_suite = unittest.TestSuite()
    add_test_cases(p_test_suite=test_suite, p_config_filename=base_test.get_config_filename())
    base_test.run_test_suite(p_test_suite=test_suite)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import abc

from python_base_app import configuration
from python_base_app import log_handling

DEFAULT_MIN_UID = 500
DEFAULT_MAX_UID = 65000
INVALID_SHELLS = ["/usr/sbin/nologin", "/bin/false"]

SYSTEM_EXCLUDE_USER_LIST = [
    'NextFreeUnixId'
]


class BaseUserHandlerConfigModel(configuration.ConfigModel):

    def __init__(self, p_section_name):
        super().__init__(p_section_name=p_section_name)
        self.min_uid = DEFAULT_MIN_UID
        self.max_uid = DEFAULT_MAX_UID

    def is_active(self):
        return self.admin_username is not None and self.admin_password is not None


class BaseUserHandler(object, metaclass=abc.ABCMeta):

    def __init__(self, p_config, p_exclude_user_list=None):

        self._logger = log_handling.get_logger(self.__class__.__name__)
        self._config = p_config
        self._exclude_user_list = p_exclude_user_list

        if self._exclude_user_list is None:
            self._exclude_user_list = []

    def is_valid_uid(self, p_uid, p_username=None, p_password=None, p_shell=None):
        if p_uid < self._config.min_uid  or p_uid > self._config.max_uid:
            return False

        if p_shell is not None:
            if p_shell in INVALID_SHELLS:
                return False

        if p_username is not None:
            if p_username in self._exclude_user_list or p_username in SYSTEM_EXCLUDE_USER_LIST:
                return False

        return True

    @abc.abstractmethod
    def list_users(self):  # pragma: no cover
        pass

    @abc.abstractmethod
    def get_uid(self, p_username):
        pass

    def authenticate(self, p_username, p_password):
        return False

    def is_admin(self, p_username):
        return False



#    Copyright (C) 2019-2021  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import unittest.mock

from python_base_app import configuration
from python_base_app import unix_user_handler
from python_base_app.test import base_test

SEARCH_BASE_DN = "dc=mydomain,dc=com"

ADMIN_USER = "admin"
ADMIN_PASSWORD = "hello!"

USER_1_UID_NUMBER = "1000"
USER_1_UID = "admin"
USER_1_PASSWORD = "$ecret"

USER_2_UID_NUMBER = "1001"
USER_2_UID = "non-admin"
USER_2_PASSWORD = "0ther$ecret"

USER_3_UID_NUMBER = "100"
USER_3_UID = "special-user"
USER_3_PASSWORD = "0ther$ecret"

CONFIG_USER_LIST = USER_1_UID + ":" + USER_1_UID_NUMBER + "," + USER_2_UID + ":" + USER_2_UID_NUMBER


class FakePasswordEntry(object):

    def __init__(self, p_uid, p_username, p_password, p_shell):
        self.pw_uid = p_uid
        self.pw_name = p_username
        self.pw_passwd = p_password
        self.pw_shell = p_shell


FAKE_USER_MAP = {
    USER_1_UID: FakePasswordEntry(p_uid=int(USER_1_UID_NUMBER), p_username=USER_1_UID, p_password=USER_1_PASSWORD,
                                  p_shell="/bin/bash"),
    USER_2_UID: FakePasswordEntry(p_uid=int(USER_2_UID_NUMBER), p_username=USER_2_UID, p_password=USER_2_PASSWORD,
                                  p_shell="/bin/bash"),
    USER_3_UID: FakePasswordEntry(p_uid=int(USER_3_UID_NUMBER), p_username=USER_3_UID, p_password=USER_3_PASSWORD,
                                  p_shell="/bin/bash"),
}


def fake_getpwall():
    return FAKE_USER_MAP.values()


def fake_getpwnam(p_username):
    return FAKE_USER_MAP.get(p_username)


class TestUnixUserHandler(base_test.BaseTestCase):

    @classmethod
    def create_dummy_unix_user_handler(cls):
        config = unix_user_handler.UnixUserHandlerConfigModel()
        config.admin_username = ADMIN_USER
        config.admin_password = ADMIN_PASSWORD
        config.user_list = CONFIG_USER_LIST

        return unix_user_handler.UnixUserHandler(p_config=config)

    # See https://pypi.org/project/fakeunix/
    def setUp(self):
        # Patch where the unix library is used:
        self.getpwall_patcher = unittest.mock.patch('pwd.getpwall')
        self.mock_getpwall = self.getpwall_patcher.start()
        self.mock_getpwall.return_value = fake_getpwall()

        self.getpwnam_patcher = unittest.mock.patch('pwd.getpwnam')
        self.mock_getpwnam = self.getpwnam_patcher.start()
        self.mock_getpwnam.side_effect = fake_getpwnam

    def tearDown(self):
        self.mock_getpwall.stop()
        self.mock_getpwnam.stop()

    @classmethod
    def get_config(cls, p_user_list=None):
        config = unix_user_handler.UnixUserHandlerConfigModel()

        config.admin_username = USER_1_UID
        config.admin_password = USER_1_PASSWORD
        config.user_list = p_user_list

        return config

    def test_inactive_config(self):
        config = unix_user_handler.UnixUserHandlerConfigModel()
        self.assertFalse(config.is_active())

    def test_active_config(self):
        config = self.get_config()
        self.assertTrue(config.is_active())

    def test_instantiation(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config())

        self.assertIsNotNone(handler)

    def test_instantiation_invalid_user_list(self):
        config = self.get_config(p_user_list="xxx:yyy")

        with self.assertRaises(configuration.ConfigurationException):
            unix_user_handler.UnixUserHandler(p_config=config)

    def test_list_users(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        users = handler.list_users()

        self.assertIsNotNone(users)
        self.assertEqual(len(users), 2)
        self.assertIn(USER_1_UID, users)
        self.assertIn(USER_2_UID, users)

    def test_list_users_without_user_list(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config())

        users = handler.list_users()

        self.assertIsNotNone(users)
        self.assertEqual(len(users), 2)
        self.assertIn(USER_1_UID, users)
        self.assertIn(USER_2_UID, users)

    def test_get_valid_uid(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertEqual(int(USER_1_UID_NUMBER), handler.get_uid(p_username=USER_1_UID))
        self.assertEqual(int(USER_2_UID_NUMBER), handler.get_uid(p_username=USER_2_UID))

    def test_get_valid_uid_without_user_list(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config())

        self.assertEqual(int(USER_1_UID_NUMBER), handler.get_uid(p_username=USER_1_UID))
        self.assertEqual(int(USER_2_UID_NUMBER), handler.get_uid(p_username=USER_2_UID))

    def test_get_invalid_uid(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertIsNone(handler.get_uid(p_username="xxx"))

    def test_get_invalid_uid_without_user_list(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config())

        self.assertIsNone(handler.get_uid(p_username="xxx"))

    def test_authenticate_non_admin_user(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertFalse(handler.authenticate(p_username=USER_2_UID, p_password=USER_2_PASSWORD))

    def test_authenticate_admin_valid_password(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertTrue(handler.authenticate(p_username=USER_1_UID, p_password=USER_1_PASSWORD))

    def test_authenticate_admin_invalid_password(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertFalse(handler.authenticate(p_username=USER_1_UID, p_password=USER_1_PASSWORD + "x"))

    def test_is_valid_admin(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertTrue(handler.is_admin(p_username=USER_1_UID))

    def test_is_invalid_admin(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertFalse(handler.is_admin(p_username=USER_2_UID))

    def test_is_invalid_uid_super_branch(self):
        handler = unix_user_handler.UnixUserHandler(p_config=self.get_config(p_user_list=CONFIG_USER_LIST))

        self.assertFalse(handler.is_valid_uid(p_uid=100, p_username="xxx"))



# -*- coding: utf-8 -*-

#    Copyright (C) 2021  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os

import requests

from python_base_app import base_app
from python_base_app import base_web_server
from python_base_app.test import base_test


class TestBaseWebServer(base_test.BaseTestCase):

    @classmethod
    def default_server(cls):

        config = base_web_server.BaseWebServerConfigModel()
        config.port = int(os.getenv("PING_API_SERVER_PORT", "6660"))
        server = base_web_server.BaseWebServer(p_config=config, p_name="BaseWebServer",
                                               p_package_name=base_app.PACKAGE_NAME)

        return server

    def test_api_health(self):

        default_server = None

        try:
            default_server = self.default_server()

            default_server.start_server()

            url = "http://localhost:{port}/health"
            r = requests.get(url.format(port=default_server._config.port))

            self.assertIsNotNone(r.text)
            self.assertEqual(r.text, "ok")

        except Exception as e:
            raise e

        finally:
            if default_server is not None:
                default_server.stop_server()
                default_server.destroy()



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

class MovingAverage(object):

    def __init__(self, p_sample_size):

        self._sample_size = p_sample_size
        self._pointer = 0
        self._values = []
        self._sum = 0.0
        self._latest_value = None

    def add_value(self, p_value):

        if len(self._values) < self._sample_size:
            self._values.append(p_value)

        else:
            self._sum = self._sum - self._values[self._pointer]
            self._values[self._pointer] = p_value
            self._pointer = self._pointer + 1
            if self._pointer == self._sample_size:
                self._pointer = 0

        self._sum = self._sum + p_value
        self._latest_value = p_value

    def get_latest_value(self):
        return self._latest_value


    def get_value(self, p_default=None):

        if len(self._values) == 0:
            return p_default

        else:
            return 1.0 * self._sum / len(self._values)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import gettext


class LocaleHelper(object):

    def __init__(self, p_locale_dir=None, p_locale_selector=None):

        self._locale_dir = p_locale_dir
        self._locale_selector = p_locale_selector
        self._langs = {}
        self._chained_helper = None

        if self._locale_selector is None:
            self._locale_selector = lambda: "en"

    def chain_helper(self, p_helper):

        self._chained_helper = p_helper

    def gettext(self, p_text):
        current_locale = self.locale

        if current_locale is None:
            current_locale = "en"

        gettext_func = self._langs.get(current_locale)

        if gettext_func is None:
            gettext_func = gettext.translation("messages", localedir=self._locale_dir,
                                               languages=[current_locale], fallback=True)
            self._langs[current_locale] = gettext_func

        if self._chained_helper is None:
            return gettext_func.gettext(p_text)

        else:
            return self._chained_helper.gettext(gettext_func.gettext(p_text))

    @property
    def locale_selector(self):
        return self._locale_selector

    @property
    def locale(self):
        return self.locale_selector()



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/little_brother
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import threading

from python_base_app import base_audio_player
from python_base_app import configuration


class PygletAudioPlayer(base_audio_player.BaseAudioPlayer):

    def __init__(self):

        super().__init__()

        self._player_thread = None

        try:
            import pyglet

            self._pyglet = pyglet
            run_method = self._pyglet.app.run
            self._pyglet.clock.schedule(lambda dt : 1)
            self._player_thread = threading.Thread(target=run_method)
            self._player_thread.start()

        except Exception:
            fmt = "Cannot load module 'pyglet'"
            self._logger.error(fmt)
            raise configuration.ConfigurationException(fmt)

        self._logger.info("audio player 'pyglet' loaded")

    def play_audio_file(self, p_audio_filename):  # pragma: no cover
        audio_file = self._pyglet.media.load(p_audio_filename)
        audio_file.play()

    def stop(self):
        if self._player_thread is not None:
            self._logger.info("audio player 'pyglet' deactivated")
            self._pyglet.app.exit()
            self._player_thread.join()



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import unittest
import os.path

from python_base_app import configuration
from python_base_app.test import base_test

SECTION_NAME = "MySection"
INT_VALUE = 123
NEW_INT_VALUE = 456
STRING_VALUE = "Hello"
BOOLEAN_VALUE = True


class SomeTestConfigModel(configuration.ConfigModel):

    def __init__(self):
        super(SomeTestConfigModel, self).__init__(p_section_name=SECTION_NAME)

        self.string = STRING_VALUE

        self.int = INT_VALUE
        self.none_int = configuration.NONE_INTEGER

        self.bool = BOOLEAN_VALUE
        self.none_bool = configuration.NONE_BOOLEAN
        
        self.empty_int_array = [configuration.NONE_INTEGER]
        self.int_array = [1, 2]

        self.empty_string_array = [configuration.NONE_STRING]
        self.string_array = ["somebody", "in", "there"]


class TestConfiguration(base_test.BaseTestCase):

    def test_configuration_int(self):

        model = SomeTestConfigModel()

        self.assertEqual(model.get_option_type("none_int"), "int")
        self.assertIsNone(model.none_int)

        self.assertEqual(model.get_option_type("int"), "int")
        self.assertEqual(model.int, INT_VALUE)

    def test_configuration_boolean(self):

        model = SomeTestConfigModel()

        self.assertEqual(model.get_option_type("none_bool"), "bool")
        self.assertIsNone(model.none_bool)

        self.assertEqual(model.get_option_type("bool"), "bool")
        self.assertEqual(model.bool, BOOLEAN_VALUE)

    def test_configuration_unknown_option(self):

        model = SomeTestConfigModel()
        exception = None

        try:
            model.some_option

        except Exception as e:
            exception = e

        self.assertIsInstance(exception, AttributeError)
        
        
    def test_valid_int_array(self):
        
        model = SomeTestConfigModel()

        self.assertIsNotNone(model.int_array)
        self.assertEqual(2, len(model.int_array))
        self.assertEqual(1, model.int_array[0])
        self.assertEqual(2, model.int_array[1])

    def test_valid_empty_int_array(self):

        model = SomeTestConfigModel()

        attr = model.none_int
        attr = model.empty_int_array
        self.assertIsNotNone(attr)
        self.assertEqual(0, len(attr))

    def test_valid_string_array(self):

        model = SomeTestConfigModel()

        self.assertIsNotNone(model.string_array)
        self.assertEqual(3, len(model.string_array))
        self.assertEqual("somebody", model.string_array[0])
        self.assertEqual("in", model.string_array[1])
        self.assertEqual("there", model.string_array[2])

    def test_valid_empty_string_array(self):

        model = SomeTestConfigModel()

        self.assertIsNotNone(model.empty_string_array)
        self.assertEqual(0, len(model.empty_string_array))

    def test_invalid_int(self):

        model = SomeTestConfigModel()

        config = configuration.Configuration()
        config.add_section(p_section=model)
        exception = None

        try:
            config.set_config_value(p_section_name=SECTION_NAME, p_option="int", p_option_value="hallo")

        except Exception as e:
            exception = e

        self.assertIsInstance(exception, configuration.ConfigurationException)
        self.assertIn("Invalid numerical value", str(exception))

    def test_invalid_none_int(self):

        model = SomeTestConfigModel()

        config = configuration.Configuration()
        config.add_section(p_section=model)
        exception = None

        try:
            config.set_config_value(p_section_name=SECTION_NAME, p_option="none_int", p_option_value="hallo")

        except Exception as e:
            exception = e

        self.assertIsInstance(exception, configuration.ConfigurationException)
        self.assertIn("Invalid numerical value", str(exception))

    def test_invalid_boolean(self):

        model = SomeTestConfigModel()

        config = configuration.Configuration()
        config.add_section(p_section=model)
        exception = None

        try:
            config.set_config_value(p_section_name=SECTION_NAME, p_option="bool", p_option_value="123")

        except Exception as e:
            exception = e

        self.assertIsInstance(exception, configuration.ConfigurationException)
        self.assertIn("Invalid Boolean value", str(exception))

    def test_invalid_none_boolean(self):

        model = SomeTestConfigModel()

        config = configuration.Configuration()
        config.add_section(p_section=model)
        exception = None

        try:
            config.set_config_value(p_section_name=SECTION_NAME, p_option="none_bool", p_option_value="123")

        except Exception as e:
            exception = e

        self.assertIsInstance(exception, configuration.ConfigurationException)
        self.assertIn("Invalid Boolean value", str(exception))

    def test_valid_boolean_values(self):

        model = SomeTestConfigModel()

        config = configuration.Configuration()
        config.add_section(p_section=model)

        for value in configuration.VALID_BOOLEAN_FALSE_VALUES:
            config.set_config_value(p_section_name=SECTION_NAME, p_option="bool", p_option_value=value)
            self.assertFalse(model.bool)

        for value in configuration.VALID_BOOLEAN_TRUE_VALUES:
            config.set_config_value(p_section_name=SECTION_NAME, p_option="bool", p_option_value=value)
            self.assertTrue(model.bool)


    def test_load_configuration(self):

        config = configuration.Configuration()
        model = SomeTestConfigModel()
        config.add_section(p_section=model)
        config_filename = os.path.join(self.get_test_data_path(), "test.config")
        config.read_config_file(p_filename=config_filename)

        self.assertEqual(234, model.int)
        self.assertEqual("There!", model.string)
        self.assertFalse(model.bool)

        self.assertEqual(len(model.int_array), 3)
        self.assertEqual(model.int_array[0], 1)
        self.assertEqual(model.int_array[1], 2)
        self.assertEqual(model.int_array[2], 3)

        self.assertEqual(len(model.empty_int_array), 3)
        self.assertEqual(model.empty_int_array[0], 4)
        self.assertEqual(model.empty_int_array[1], 5)
        self.assertEqual(model.empty_int_array[2], 6)

        self.assertEqual(len(model.string_array), 4)
        self.assertEqual(model.string_array[0], "somebody")
        self.assertEqual(model.string_array[1], "in")
        self.assertEqual(model.string_array[2], "there")
        self.assertEqual(model.string_array[3], "?")

        self.assertEqual(len(model.empty_string_array), 3)
        self.assertEqual(model.empty_string_array[0], "one")
        self.assertEqual(model.empty_string_array[1], "two")
        self.assertEqual(model.empty_string_array[2], "three")


    def test_override_by_command_line_options(self):

        config = configuration.Configuration()
        model = SomeTestConfigModel()
        config.add_section(p_section=model)

        self.assertEqual(model.int, INT_VALUE)

        parameters = [
            "{section}.int={value}".format(section=SECTION_NAME, value=NEW_INT_VALUE)
        ]

        config.read_command_line_parameters(parameters)

        self.assertEqual(model.int, NEW_INT_VALUE)

    def test_override_by_environment(self):

        config = configuration.Configuration()
        model = SomeTestConfigModel()
        config.add_section(p_section=model)

        self.assertEqual(model.int, INT_VALUE)

        environment = {
            "{section}__int".format(section=SECTION_NAME): str(NEW_INT_VALUE)
        }

        config.read_environment_parameters(p_environment_dict=environment)

        self.assertEqual(model.int, NEW_INT_VALUE)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import logging.handlers
import os
from os.path import join

g_logging_started = False

LOG_BACKUP_COUNT = 10
LOG_MAX_BYTES = 1000000

class LogDataHandler(object):

    def get_log_username(self):
        return None

    def get_log_client_ip(self):
        return None


g_log_data_handler = LogDataHandler()


# Siehe https://docs.python.org/3/howto/logging-cookbook.html#context-info
class LogFilter(logging.Filter):

    def filter(self, record):
        global g_log_data_handler

        username = g_log_data_handler.get_log_username()
        client_ip = g_log_data_handler.get_log_client_ip()

        record.raw_username = username
        record.raw_client_ip = client_ip

        record.username = "login=%s - " % username if username is not None else ""
        record.client_ip = "client-ip=%s - " % client_ip if client_ip is not None else ""
        return True


g_log_filter = LogFilter()


def register_log_data_handler(p_log_data_handler):
    global g_log_data_handler

    g_log_data_handler = p_log_data_handler


def get_log_level_by_name(p_log_level_name):
    return getattr(logging, p_log_level_name, None)


def start_logging(p_log_dir=None, p_log_file=None, p_level=logging.DEBUG, p_use_filter=True):
    global g_logging_started
    global g_log_data_handler
    global g_log_filter

    if (not g_logging_started):

        g_logging_started = True
        logger = get_logger()
        logger.setLevel(p_level)

        if p_log_dir is not None:

            log_filename = join(p_log_dir, p_log_file)
            h = logging.handlers.RotatingFileHandler(log_filename, backupCount=LOG_BACKUP_COUNT, maxBytes=LOG_MAX_BYTES,
                                                     encoding="UTF-8")

        else:

            h = logging.StreamHandler()

        # create formatter and add it to the handlers
        if p_use_filter:
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(client_ip)s%(username)s%(message)s')

        else:
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        h.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(h)
        add_default_filter_to_logger(logger)

        fmt = "Started logging in CWD={cwd} using module {module_name}"
        logger.info(fmt.format(cwd=os.getcwd(), module_name=__name__))


def add_default_filter_to_logger(p_logger):
    global g_log_filter

    p_logger.addFilter(g_log_filter)


def set_level(p_log_level):
    msg = "Set logging level to {level}"
    get_logger().info(msg.format(level=p_log_level))

    get_logger().setLevel(get_log_level_by_name(p_log_level_name=p_log_level))

    for h in get_logger().handlers:
        h.setLevel(get_log_level_by_name(p_log_level_name=p_log_level))


def add_default_filter_to_logger_by_name(p_name):
    logger = logging.getLogger(p_name)
    add_default_filter_to_logger(p_logger=logger)


def get_logger(p_name=None):
    logger = logging.getLogger(p_name)
    add_default_filter_to_logger(p_logger=logger)
    return logger


def register_handler(p_log_handler):
    logger = get_logger()
    logger.addHandler(p_log_handler)



# -*- coding: utf-8 -*-
'''
***
Modified generic daemon class
***

Author:         http://www.jejik.com/articles/2007/02/
                        a_simple_unix_linux_daemon_in_python/www.boxedice.com

License:        http://creativecommons.org/licenses/by-sa/3.0/

Changes:        23rd Jan 2009 (David Mytton <david@boxedice.com>)
                - Replaced hard coded '/dev/null in __init__ with os.devnull
                - Added OS check to conditionally remove code that doesn't
                  work on OS X
                - Added output to console on completion
                - Tidied up formatting
                11th Mar 2009 (David Mytton <david@boxedice.com>)
                - Fixed problem with daemon exiting on Python 2.4
                  (before SystemExit was part of the Exception base)
                13th Aug 2010 (David Mytton <david@boxedice.com>
                - Fixed unhandled exception if PID file is empty
'''

# Core modules
from __future__ import print_function

import atexit
import errno
import os
import signal
import sys
import time

from python_base_app import log_handling
from python_base_app import tools


class Daemon(object):
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """

    def __init__(self, pidfile, stdin=os.devnull,
                 stdout=os.devnull, stderr=os.devnull,
                 home_dir='.', umask=0o22, verbose=1,
                 use_gevent=False, use_eventlet=False):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile
        self.home_dir = home_dir
        self.verbose = verbose
        self.umask = umask
        self.daemon_alive = True
        self.use_gevent = use_gevent
        self.use_eventlet = use_eventlet

        self._logger = log_handling.get_logger(self.__class__.__name__)

    def log(self, *args):

        self._logger.info(*args)
        # if self.verbose >= 1:
        #    print(*args)

    def daemonize(self):
        """
        Do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        #         if self.use_eventlet:
        #             import eventlet.tpool
        #             eventlet.tpool.killall()
        try:
            pid = os.fork()
            if pid > 0:
                # Exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write(
                "fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # Decouple from parent environment
        os.chdir(self.home_dir)
        os.setsid()
        os.umask(self.umask)

        # Do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # Exit from second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write(
                "fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        if sys.platform != 'darwin':  # This block breaks on OS X
            # Redirect standard file descriptors
            sys.stdout.flush()
            sys.stderr.flush()
            si = open(self.stdin, 'r')
            so = open(self.stdout, 'a+')
            if self.stderr:
                try:
                    se = open(self.stderr, 'a+', 0)
                except ValueError:
                    # Python 3 can't have unbuffered text I/O
                    se = open(self.stderr, 'a+', 1)
            else:
                se = so
            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())

        #         def sigtermhandler(signum, frame):
        #             self.daemon_alive = False
        #             sys.exit()
        #
        #         if self.use_gevent:
        #             import gevent
        #             gevent.reinit()
        #             gevent.signal(signal.SIGTERM, sigtermhandler, signal.SIGTERM, None)
        #             gevent.signal(signal.SIGINT, sigtermhandler, signal.SIGINT, None)
        #         else:
        #             signal.signal(signal.SIGTERM, sigtermhandler)
        #             signal.signal(signal.SIGINT, sigtermhandler)

        self.log("Started daemon.")

        # Write pidfile
        atexit.register(
            self.delpid)  # Make sure pid file is removed if we quit
        pid = str(os.getpid())
        open(self.pidfile, 'w+').write("%s\n" % pid)

    def delpid(self):
        try:
            # the process may fork itself again
            pid = int(open(self.pidfile, 'r').read().strip())
            if pid == os.getpid():
                os.remove(self.pidfile)
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise

    def start(self, *args, **kwargs):
        """
        Start the daemon
        """

        self.log("Starting daemon...")

        # Check for a pidfile to see if the daemon already runs
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None
        except SystemExit:
            pid = None

        if pid:
            message = "pidfile %s already exists. Is it already running?\n"
            sys.stderr.write(message % self.pidfile)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run(*args, **kwargs)

    def stop(self):
        """
        Stop the daemon
        """

        if self.verbose >= 1:
            self.log("Stopping daemon...")

        # Get the pid from the pidfile
        pid = self.get_pid()

        if not pid:
            message = "pidfile %s does not exist. Not running?\n"
            sys.stderr.write(message % self.pidfile)

            # Just to be sure. A ValueError might occur if the PID file is
            # empty but does actually exist
            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)

            return  # Not an error in a restart

        # Try killing the daemon process
        try:
            i = 0
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
                i = i + 1

                if not tools.is_windows():
                    if i % 10 == 0:
                        os.kill(pid, signal.SIGHUP)
        except OSError as err:
            if err.errno == errno.ESRCH:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err))
                sys.exit(1)

        self.log("Daemon stopped")

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def get_pid(self):
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None
        except SystemExit:
            pid = None
        return pid

    def is_running(self):
        pid = self.get_pid()

        if pid is None:
            self.log('Process is stopped')
            return False
        elif os.path.exists('/proc/%d' % pid):
            self.log('Process (pid %d) is running...' % pid)
            return True
        else:
            self.log('Process (pid %d) is killed' % pid)
            return False

    def run(self):
        """
        You should override this method when you subclass Daemon.
        It will be called after the process has been
        daemonized by start() or restart().
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

#    Copyright (C) 2019-2022  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import argparse
import collections
import copy
import logging
import os
import os.path
import stat
import subprocess
import re

import jinja2
import sys
import time

import python_base_app
from python_base_app import exceptions
from python_base_app import log_handling
from python_base_app import tools

MODULE_NAME = "base_ci_toolbox"

STAGE_BUILD_PACKAGE = "BUILD"
STAGE_BUILD_DOCKER_IMAGES = "BUILD_DOCKER_IMAGES"
STAGE_INSTALL = "INSTALL"
STAGE_INSTALL_PYPI_PACKAGE = "INSTALL-PYPI-PACKAGE"
STAGE_PUBLISH_PACKAGE = "PUBLISH-PACKAGE"
STAGE_PUBLISH_PYPI_PACKAGE = "PUBLISH-PYPI-PACKAGE"
STAGE_TEST = "TEST"
STAGE_ANALYZE = "ANALYZE"
STAGE_TEARDOWN = "TEARDOWN"
STAGE_PREPARE = "PREPARE"

GITLAB_CI_FILE = '.gitlab-ci.yml'
GITLAB_CI_TEMPLATE = 'gitlab-ci.template.yml'

GIT_PY_FILE_PATH = '{module_name}/git.py'
GIT_PY_TEMPLATE = 'git.template.py'

CIRCLE_CI_FILE = '.circleci/config.yml'
CIRCLE_CI_TEMPLATE = 'circle-ci.template.yml'

PYCOVERAGE_FILE_PATH = '.coveragerc'
PYCOVERAGE_TEMPLATE = 'coveragerc.template'

CODACY_FILE_PATH = '.codacy.yml'
CODACY_TEMPLATE = 'codacy.template.yml'

DEBIAN_CONTROL_FILE_PATH = '{debian_build_dir}/DEBIAN/control'
DEBIAN_CONTROL_TEMPLATE = 'debian_control.template.conf'

DEBIAN_POSTINST_FILE_PATH = '{debian_build_dir}/DEBIAN/postinst'
DEBIAN_POSTINST_TEMPLATE = 'debian_postinst.template.sh'

GENERIC_INSTALLATION_SCRIPT_FILE_PATH = '{bin_dir}/generic-install.sh'
GENERIC_INSTALLATION_SCRIPT_TEMPLATE = 'debian_postinst.template.sh'

MAKE_DEBIAN_PACKAGE_SCRIPT_FILE_PATH = '{bin_dir}/make-debian-package.sh'
MAKE_DEBIAN_PACKAGE_SCRIPT_TEMPLATE = 'make-debian-package.template.sh'

BUILD_DOCKER_IMAGE_SCRIPT_FILE_PATH = '{bin_dir}/build-docker-images.sh'
BUILD_DOCKER_IMAGE_SCRIPT_TEMPLATE = 'build-docker-images.template.sh'

INSTALL_DEBIAN_PACKAGE_SCRIPT_FILE_PATH = '{bin_dir}/install-debian-package.sh'
INSTALL_DEBIAN_PACKAGE_SCRIPT_TEMPLATE = 'install-debian-package.template.sh'

INSTALL_PYPI_PACKAGE_SCRIPT_FILE_PATH = '{bin_dir}/install-pypi-package.sh'
INSTALL_PYPI_PACKAGE_SCRIPT_TEMPLATE = 'install-pypi-package.template.sh'

PUBLISH_DEBIAN_PACKAGE_SCRIPT_FILE_PATH = '{bin_dir}/publish-debian-package.sh'
PUBLISH_DEBIAN_PACKAGE_SCRIPT_TEMPLATE = 'publish-debian-package.template.sh'

PUBLISH_PYPI_PACKAGE_SCRIPT_FILE_PATH = '{bin_dir}/publish-pypi-package.sh'
PUBLISH_PYPI_PACKAGE_SCRIPT_TEMPLATE = 'publish-pypi-package.template.sh'

PIP3_SCRIPT_FILE = '{bin_dir}/pip3.sh'
PIP3_SCRIPT_TEMPLATE = 'pip3.template.sh'

TEST_APP_SCRIPT_FILE_PATH = '{bin_dir}/test-app.sh'
TEST_APP_SCRIPT_TEMPLATE = 'test-app.template.sh'

ANALYZE_APP_SCRIPT_FILE_PATH = '{bin_dir}/analyze-app.sh'
ANALYZE_APP_SCRIPT_TEMPLATE = 'analyze-app.template.sh'

BEACON_INTERVAL = 60  # seconds

VarStatus = collections.namedtuple("VarInfo", "source_name description target_name")

PREDEFINED_ENV_VARIABLES = [
    VarStatus('FORCED_GIT_BRANCH', "Forced Git branch name (possibly overriding CI/CD branch names)", "GIT_BRANCH"),
    VarStatus('CIRCLE_BRANCH', "Circle CI Git branch name", "GIT_BRANCH"),
    VarStatus('CI_COMMIT_BRANCH', "GitLab CI/CD Git branch name", "GIT_BRANCH"),
    VarStatus('DOCKER_REGISTRY_HOST_NAME', "Docker registry host name", "docker_registry"),
    VarStatus('DOCKER_REGISTRY_ORG_UNIT', "Docker registry organisational unit", "docker_registry_org_unit"),
    VarStatus('DOCKER_REGISTRY_USER', "Docker registry login user", "docker_registry_user"),
    VarStatus('MAX_CPUS', "Maximum number of parallel CPUs/threads to be used during build", "max_cpus"),
]

DEFAULT_PYPI_REPOSITORY = "https://upload.pypi.org/legacy/"
DEFAULT_TEST_PYPI_REPOSITORY = "https://test.pypi.org/legacy/"

DEFAULT_PYPI_URL_ENV_NAME = "PYPI_API_URL"
DEFAULT_TEST_PYPI_URL_ENV_NAME = "TEST_PYPI_API_URL"

DEFAULT_TEST_PYPI_EXTRA_INDEX_ENV_NAME = "TEST_PYPI_EXTRA_INDEX"
DEFAULT_TEST_PYPI_DELETE_PACKAGE_ENV_NAME = "TEST_PYPI_DELETE_PACKAGE"

DEFAULT_PYPI_TOKEN_ENV_NAME = "PYPI_API_TOKEN"
DEFAULT_PYPI_USER_ENV_NAME = "PYPI_API_USER"

predefined_env_variables = None

default_setup = {
    "docker_image_make_package": "marcusrickert/docker-python-app:latest",
    "docker_image_test": "marcusrickert/docker-python-app:latest",
    "docker_image_docker": "marcusrickert/docker-docker-ci:release-0.9.1",
    "docker_image_analyze": "marcusrickert/docker-python-app:latest",
    "docker_image_owasp": "registry.gitlab.com/gitlab-ci-utils/docker-dependency-check:latest",
    "owasp_additional_params": "--enableExperimental",
    "ci_toolbox_script": "ci_toolbox.py",
    "ci_pip_dependencies": [ ],
    "ci_stage_build_package": STAGE_BUILD_PACKAGE,
    "ci_stage_build_docker_images": STAGE_BUILD_DOCKER_IMAGES,
    "ci_stage_install": STAGE_INSTALL,
    "ci_stage_install_pypi_package": STAGE_INSTALL_PYPI_PACKAGE,
    "ci_stage_test": STAGE_TEST,
    "ci_stage_analyze": STAGE_ANALYZE,
    "ci_stage_teardown": STAGE_TEARDOWN,
    "ci_stage_publish_package": STAGE_PUBLISH_PACKAGE,
    "ci_stage_publish_pypi_package": STAGE_PUBLISH_PYPI_PACKAGE,
    "ci_stage_build_pip_dependencies": [ "babel" ],
    "ci_stage_publish_pip_package_pip_dependencies": [ "twine" ],
    "owasp": False,
    "require_teardown": False,
    "bin_dir": "bin",
    "test_dir": "test",
    "run_test_suite": "run_{module_name}_test_suite.py",
    "run_test_suite_no_venv": "run_{module_name}_test_suite_no_venv.py",
    "contrib_dir": "contrib",
    "rel_tmp_dir": "tmp",
    "rel_etc_dir": "etc/{name}",
    "rel_log_dir": "var/log/{name}",
    "rel_spool_dir": "var/spool/{name}",
    "rel_lib_dir": "var/lib/{name}",
    "rel_virtual_env_dir": "var/lib/{name}/virtualenv",
    "rel_systemd_dir": "lib/systemd/system",
    "rel_tmpfile_dir": "usr/lib/tmpfiles.d",
    "rel_sudoers_dir": "etc/sudoers.d",
    "rel_apparmor_dir": "etc/apparmor.d",
    "git_metadata_file": "{module_name}/git_metadata.py",
    "user": "{name}",
    "group": "{name}",
    "user_group_mappings": [],
    "create_usr": False,
    "create_group": False,
    "deploy_systemd_service": False,
    "deploy_tmpfile_conf": False,
    "deploy_sudoers_file": False,
    "deploy_apparmor_file": False,
    "version": "0.1",
    "target_alembic_version": None,
    "build_debian_package": True,
    "debian_build_dir": "debian",
    "debian_package_name": "{name}",
    "debian_package_revision": "1",
    "debian_package_section": "base",
    "debian_package_priority": "optional",
    "debian_package_architecture": "amd64",
    "debian_dependencies": [],
    "debian_extra_files": [],
    "debian_templates": [],
    "install_requires": [],
    "contributing_setups": [],
    "publish_debian_package": [],
    "build_pypi_package": False,
    "publish_pypi_package": {},
    "extra_pypi_indexes": {},
    "publish_docker_images": [],
    "publish_latest_docker_image": "",
    "docker_registry": "docker.io",
    "docker_registry_user": "[DOCKER_REGISTRY_USER_NOT_SET]",
    "docker_registry_org_unit": "[DOCKER_REGISTRY_ORG_UNIT_NOT_SET]",
    "docker_context_dir": "docker",
    "docker_contexts": [],
    "babel_rel_directory": None,
    "generate_generic_install": False,
    "max_cpus": None,
    "analyze": False,
    "analyze_extra_coverage_exclusions": None,
    "analyze_extra_exclusions": None,
    "script_timeout": 600,
    "activate_xvfb_for_tests" : False,
}

logger:logging.Logger = None


def get_module_dir(p_module):
    return os.path.dirname(p_module.__file__)


def get_python_package_name(p_var):
    return "{name}-{version}.tar.gz".format(**(p_var["setup"]))


def get_predefined_environment_variables():
    global logger
    global predefined_env_variables

    if predefined_env_variables is None:
        predefined_env_variables = {}

        for var_info in PREDEFINED_ENV_VARIABLES:
            value = os.getenv(var_info.source_name)

            if value is not None:
                if var_info.target_name in predefined_env_variables:
                    msg = "New value '{value}' found in environment with value ({name}) to be used for '{target_name}' WILL BE IGNORED!"
                    logger.warning(msg.format(name=var_info.source_name, value=value, target_name=var_info.target_name))

                else:
                    msg = "{description} ({name}) found in environment with value '{value}' to be used for '{target_name}'"
                    logger.info(msg.format(name=var_info.source_name, description=var_info.description, value=value,
                                           target_name=var_info.target_name))
                    predefined_env_variables[var_info.target_name] = value

    return predefined_env_variables


def expand_vars(p_vars):
    while True:
        change_done = False

        for (key, value) in p_vars.items():
            if isinstance(value, str):
                try:
                    new_value = value.format(**p_vars)

                except Exception:
                    new_value = value

                if new_value != value:
                    change_done = True
                    p_vars[key] = new_value

        if not change_done:
            break


def get_vars(p_setup_params):
    setup = copy.copy(default_setup)
    setup.update(p_setup_params)

    setup["module_name"] = setup["name"].replace("-", "_")
    setup.update(get_predefined_environment_variables())

    expand_vars(setup)

    # store bin path so that tools can be found
    setup["python_base_app_bin_dir"] = os.path.realpath(os.path.join(os.path.dirname(__file__), "../bin"))

    return {"setup": setup}


def load_setup_module(p_dir, p_module_name):
    import importlib.util
    spec = importlib.util.spec_from_file_location(p_module_name, "{dir}/setup.py".format(dir=p_dir))
    setup_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(setup_module)
    return setup_module


def load_contributing_setup_modules(p_main_setup_module):
    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    contrib_setup_modules = []

    for contrib_module_name in var["setup"]["contributing_setups"]:
        module = load_setup_module(p_dir=os.path.join(var["setup"]["contrib_dir"], contrib_module_name),
                                   p_module_name=contrib_module_name)
        contrib_setup_modules.append(module)

    return contrib_setup_modules


def get_site_packages_dir():

    major_version = sys.version_info[0]
    minor_version = sys.version_info[1]

    if tools.is_windows():
        pattern = f".+Python\.{major_version}\.{minor_version}.+\\\\lib$"

    else:
        pattern = f".+python{major_version}\.{minor_version}/(site|dist)-packages$"

    regex = re.compile(pattern)

    site_packages_dirs = [p for p in sys.path if regex.match(p)]

    if len(site_packages_dirs) != 1:
        msg = f"Cannot determine unique site-package dir with pattern '{pattern}'! Candidates {site_packages_dirs} remaining from {sys.path}"
        raise Exception(msg)

    return site_packages_dirs[0]


def get_python_packages(p_main_setup_module, p_arguments, p_include_contrib_packages=True):
    """
        Return tuples describing name and path details of the python packages used for the application.
        
        @return Array of tuples containing
            * 0: path to the package source parent dir
            * 1: filename of the python pip package (excluding path)
            * 2: module name
            * 3: vars of the package
            * 4: variable name containing the target PyPi repository URL
            * 5: variable name containing the token for the target PyPi repository
            * 6: variable name containing the username for the target PyPi repository
            * 7: default target PyPi repository URL
            * 8: variable name for extra PyPi index repository URL
            * 9: variable name containing the URL for listing PyPi packages for subsequent deletion
            * 10: name of the PyPi package (no extension, no version)
            * 11: version of the PyPi package
    """

    if p_arguments is None:
        return []

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    if p_arguments.use_dev_dir is not None:
        contrib_dir = os.path.join(p_arguments.use_dev_dir, var["setup"]["contrib_dir"])
        app_dir = p_arguments.use_dev_dir

    else:
        contrib_dir = get_site_packages_dir()
        app_dir = get_site_packages_dir()

    branch = var["setup"].get("GIT_BRANCH")
    branch_target_rep_map = var["setup"]["publish_pypi_package"]
    branch_extra_pypi_indexes_map = var["setup"]["extra_pypi_indexes"]

    target_rep_url_env_name = None
    target_rep_token_env_name = None
    target_rep_user_env_name = None
    target_rep_extra_index_env_name = None

    if branch is not None:
        if isinstance(branch_target_rep_map, dict):
            target_rep_map_entry = branch_target_rep_map.get(branch)

            if target_rep_map_entry is not None:
                target_rep_url_env_name = target_rep_map_entry[0]
                target_rep_token_env_name = target_rep_map_entry[1]
                target_rep_user_env_name = target_rep_map_entry[2]

    if target_rep_url_env_name is None:
        target_rep_url_env_name = DEFAULT_PYPI_URL_ENV_NAME

    if target_rep_token_env_name is None:
        target_rep_token_env_name = DEFAULT_PYPI_TOKEN_ENV_NAME

    if target_rep_user_env_name is None:
        target_rep_user_env_name = DEFAULT_PYPI_USER_ENV_NAME

    if target_rep_url_env_name == DEFAULT_TEST_PYPI_URL_ENV_NAME:
        target_rep_default_url = DEFAULT_TEST_PYPI_REPOSITORY

    else:
        target_rep_default_url = DEFAULT_PYPI_REPOSITORY

    contributing_setup_modules = load_contributing_setup_modules(p_main_setup_module)

    python_packages = []

    python_packages.append((app_dir,                                                            # 0
                            get_python_package_name(p_var=var),                                 # 1
                            var["setup"]["module_name"],                                        # 2
                            var,                                                                # 3
                            target_rep_url_env_name,                                            # 4
                            target_rep_token_env_name,                                          # 5
                            target_rep_user_env_name,                                           # 6
                            target_rep_default_url,                                             # 7
                            DEFAULT_TEST_PYPI_EXTRA_INDEX_ENV_NAME,                             # 8
                            DEFAULT_TEST_PYPI_DELETE_PACKAGE_ENV_NAME,                          # 9
                            var["setup"]["name"],                                               # 10
                            var["setup"]["version"]))                                           # 11

    if p_include_contrib_packages:
        for contributing_setup_module in contributing_setup_modules:
            contrib_var = get_vars(p_setup_params=contributing_setup_module.extended_setup_params)
            module_name = contrib_var["setup"]["module_name"]

            if p_arguments.use_dev_dir is not None:
                include_path = os.path.join(contrib_dir, module_name)

            else:
                include_path = contrib_dir

            python_packages.append((include_path, get_python_package_name(p_var=contrib_var), module_name, contrib_var,
                                    target_rep_url_env_name, target_rep_token_env_name, target_rep_user_env_name,
                                    target_rep_default_url,
                                    DEFAULT_TEST_PYPI_EXTRA_INDEX_ENV_NAME,
                                    DEFAULT_TEST_PYPI_DELETE_PACKAGE_ENV_NAME,
                                    contrib_var["setup"]["name"],
                                    contrib_var["setup"]["version"]))

    return python_packages


def generate_standard_file(p_main_setup_module, p_template_env, p_file_description,
                           p_output_filename, p_template_name, p_create_directory=False,
                           p_arguments=None, p_make_executable=False):
    global logger

    fmt = "Generate {file_description} for version {version} of app '{name}'"
    logger.info(fmt.format(file_description=p_file_description, **p_main_setup_module.extended_setup_params), )

    template = p_template_env.get_template(p_template_name)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(var=get_vars(p_setup_params=p_main_setup_module.extended_setup_params),
                                  python_packages=get_python_packages(p_main_setup_module=p_main_setup_module,
                                                                      p_arguments=p_arguments))

    output_file_path = p_output_filename.format(**(var["setup"]))

    filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    if p_create_directory:
        directory = os.path.dirname(filename)
        os.makedirs(directory, exist_ok=True)

    with open(filename, "w") as f:
        f.write(output_text)

    if p_make_executable:
        os.chmod(filename, 0o755)

    fmt = "Wrote {file_description} file '{filename}', executable bit: {executable}"
    logger.info(fmt.format(filename=filename, file_description=p_file_description, executable=p_make_executable))


def generate_git_python_file(p_main_setup_module, p_template_env):
    generate_standard_file(p_main_setup_module=p_main_setup_module, p_template_env=p_template_env,
                           p_file_description="git.py",
                           p_output_filename=GIT_PY_FILE_PATH, p_template_name=GIT_PY_TEMPLATE)


def generate_circle_ci_configuration(p_main_setup_module, p_template_env, p_arguments):
    generate_standard_file(p_main_setup_module=p_main_setup_module, p_template_env=p_template_env,
                           p_file_description="Circle CI configuration",
                           p_output_filename=CIRCLE_CI_FILE, p_template_name=CIRCLE_CI_TEMPLATE,
                           p_create_directory=True, p_arguments=p_arguments)


def generate_gitlab_ci_configuration(p_main_setup_module, p_template_env, p_arguments):
    generate_standard_file(p_main_setup_module=p_main_setup_module, p_template_env=p_template_env,
                           p_file_description="GitLab CI configuration",
                           p_output_filename=GITLAB_CI_FILE, p_template_name=GITLAB_CI_TEMPLATE,
                           p_arguments=p_arguments)

def generate_pip3_script(p_main_setup_module, p_template_env, p_arguments):
    generate_standard_file(p_main_setup_module=p_main_setup_module, p_template_env=p_template_env,
                           p_file_description="pip3 script",
                           p_output_filename=PIP3_SCRIPT_FILE, p_template_name=PIP3_SCRIPT_TEMPLATE,
                           p_arguments=p_arguments, p_make_executable=True, p_create_directory=True)


def generate_codacy_configuration(p_main_setup_module, p_template_env):
    generate_standard_file(p_main_setup_module=p_main_setup_module, p_template_env=p_template_env,
                           p_file_description="Codacy configuration",
                           p_output_filename=CODACY_FILE_PATH, p_template_name=CODACY_TEMPLATE)


def generate_debian_control(p_main_setup_module, p_template_env):
    global logger

    fmt = "Generate Debian control file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(DEBIAN_CONTROL_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(var=var, depends=var["setup"]["install_requires"])

    output_file_path = DEBIAN_CONTROL_FILE_PATH.format(**(var["setup"]))

    debian_control_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(debian_control_filename), mode=0o777, exist_ok=True)

    with open(debian_control_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote Debian control file to '{filename}'"
    logger.info(fmt.format(filename=debian_control_filename))


def generate_debian_postinst(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate Debian post installation file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(DEBIAN_POSTINST_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        generic_script=False,
        user_group_mappings=var["setup"]["user_group_mappings"],
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments)
    )

    output_file_path = DEBIAN_POSTINST_FILE_PATH.format(**(var["setup"]))

    debian_postinst_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(debian_postinst_filename), mode=0o777, exist_ok=True)

    with open(debian_postinst_filename, "w") as f:
        f.write(output_text)

    os.chmod(debian_postinst_filename,
             stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP | stat.S_IROTH | stat.S_IXOTH)

    fmt = "Wrote Debian post installation file to '{filename}'"
    logger.info(fmt.format(filename=debian_postinst_filename))


def generate_generic_installation_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate generic installation file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(GENERIC_INSTALLATION_SCRIPT_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        generic_script=True,
        user_group_mappings=var["setup"]["user_group_mappings"],
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments)
    )

    output_file_path = GENERIC_INSTALLATION_SCRIPT_FILE_PATH.format(**(var["setup"]))

    debian_postinst_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(debian_postinst_filename), mode=0o777, exist_ok=True)

    with open(debian_postinst_filename, "w") as f:
        f.write(output_text)

    os.chmod(debian_postinst_filename,
             stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP | stat.S_IROTH | stat.S_IXOTH)

    fmt = "Wrote generic installation script to '{filename}'"
    logger.info(fmt.format(filename=debian_postinst_filename))


def generate_pycoveragerc(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate pycoverage configuration file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(PYCOVERAGE_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments,
                                            p_include_contrib_packages=False)
    )

    output_file_path = PYCOVERAGE_FILE_PATH.format(**(var["setup"]))

    if p_arguments.run_dir is not None:
        run_dir = p_arguments.run_dir

    else:
        run_dir = get_module_dir(p_module=p_main_setup_module)

    pycoveragerc_filename = os.path.join(run_dir, output_file_path)

    with open(pycoveragerc_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote pycoverage configuration file to '{filename}'"
    logger.info(fmt.format(filename=pycoveragerc_filename))


def generate_make_debian_package(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate make_debian_package.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(MAKE_DEBIAN_PACKAGE_SCRIPT_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments)
    )

    output_file_path = MAKE_DEBIAN_PACKAGE_SCRIPT_FILE_PATH.format(**(var["setup"]))

    make_debian_package_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(make_debian_package_script_filename), mode=0o777, exist_ok=True)

    with open(make_debian_package_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote make_debian_package.sh script file to '{filename}'"
    logger.info(fmt.format(filename=make_debian_package_script_filename))

    os.chmod(make_debian_package_script_filename,
             stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def generate_build_docker_image_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate build_docker_images.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(BUILD_DOCKER_IMAGE_SCRIPT_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments)
    )

    output_file_path = BUILD_DOCKER_IMAGE_SCRIPT_FILE_PATH.format(**(var["setup"]))

    build_docker_image_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(build_docker_image_script_filename), mode=0o777, exist_ok=True)

    with open(build_docker_image_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote build_docker_image.sh script file to '{filename}'"
    logger.info(fmt.format(filename=build_docker_image_script_filename))

    os.chmod(build_docker_image_script_filename,
             stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def generate_install_debian_package_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate install-debian-package.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(INSTALL_DEBIAN_PACKAGE_SCRIPT_TEMPLATE)
    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)
    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments)
    )
    output_file_path = INSTALL_DEBIAN_PACKAGE_SCRIPT_FILE_PATH.format(**(var["setup"]))
    install_and_test_debian_package_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module),
                                                                   output_file_path)
    os.makedirs(os.path.dirname(install_and_test_debian_package_script_filename), mode=0o777, exist_ok=True)

    with open(install_and_test_debian_package_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote install-debian-package.sh script file to '{filename}'"
    logger.info(fmt.format(filename=install_and_test_debian_package_script_filename))

    os.chmod(install_and_test_debian_package_script_filename,
             stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def generate_install_pypi_package_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate install-pypi-package.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(INSTALL_PYPI_PACKAGE_SCRIPT_TEMPLATE)
    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)
    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments)
    )
    output_file_path = INSTALL_PYPI_PACKAGE_SCRIPT_FILE_PATH.format(**(var["setup"]))
    install_and_test_pypi_package_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module),
                                                                 output_file_path)
    os.makedirs(os.path.dirname(install_and_test_pypi_package_script_filename), mode=0o777, exist_ok=True)

    with open(install_and_test_pypi_package_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote install-pypi-package.sh script file to '{filename}'"
    logger.info(fmt.format(filename=install_and_test_pypi_package_script_filename))

    os.chmod(install_and_test_pypi_package_script_filename,
             stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def generate_test_app_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate test-app.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(TEST_APP_SCRIPT_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments),
        arguments=p_arguments,
        site_packages_dir=get_site_packages_dir()
    )

    output_file_path = TEST_APP_SCRIPT_FILE_PATH.format(**(var["setup"]))

    test_app_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(test_app_script_filename), mode=0o777, exist_ok=True)

    with open(test_app_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote test-app.sh script file to '{filename}'"
    logger.info(fmt.format(filename=test_app_script_filename))

    os.chmod(test_app_script_filename, stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)

def generate_analyze_app_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate analyze-app.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(ANALYZE_APP_SCRIPT_TEMPLATE)

    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments),
        arguments=p_arguments,
        site_packages_dir=get_site_packages_dir()
    )

    output_file_path = ANALYZE_APP_SCRIPT_FILE_PATH.format(**(var["setup"]))

    analyze_app_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    os.makedirs(os.path.dirname(analyze_app_script_filename), mode=0o777, exist_ok=True)

    with open(analyze_app_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote analyze-app.sh script file to '{filename}'"
    logger.info(fmt.format(filename=analyze_app_script_filename))

    os.chmod(analyze_app_script_filename, stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def generate_publish_debian_package_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate publish-debian-package.template.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(PUBLISH_DEBIAN_PACKAGE_SCRIPT_TEMPLATE)
    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)
    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments),
        arguments=p_arguments,
        site_packages_dir=get_site_packages_dir()
    )
    output_file_path = PUBLISH_DEBIAN_PACKAGE_SCRIPT_FILE_PATH.format(**(var["setup"]))
    publish_debian_package_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module),
                                                          output_file_path)
    os.makedirs(os.path.dirname(publish_debian_package_script_filename), mode=0o777, exist_ok=True)

    with open(publish_debian_package_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote publish_debian_package.sh script file to '{filename}'"
    logger.info(fmt.format(filename=publish_debian_package_script_filename))

    os.chmod(publish_debian_package_script_filename,
             stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def generate_publish_pypi_package_script(p_main_setup_module, p_template_env, p_arguments):
    global logger

    fmt = "Generate publish-pypi-package.template.sh script file for version {version} of app '{name}'"
    logger.info(fmt.format(**p_main_setup_module.extended_setup_params))

    template = p_template_env.get_template(PUBLISH_PYPI_PACKAGE_SCRIPT_TEMPLATE)
    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)
    output_text = template.render(
        var=var,
        python_packages=get_python_packages(p_main_setup_module=p_main_setup_module, p_arguments=p_arguments,
                                            p_include_contrib_packages=False),
        arguments=p_arguments,
        site_packages_dir=get_site_packages_dir()
    )
    output_file_path = PUBLISH_PYPI_PACKAGE_SCRIPT_FILE_PATH.format(**(var["setup"]))
    publish_pypi_package_script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)
    os.makedirs(os.path.dirname(publish_pypi_package_script_filename), mode=0o777, exist_ok=True)

    with open(publish_pypi_package_script_filename, "w") as f:
        f.write(output_text)

    fmt = "Wrote publish_pypi_package.sh script file to '{filename}'"
    logger.info(fmt.format(filename=publish_pypi_package_script_filename))

    os.chmod(publish_pypi_package_script_filename,
             stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXGRP | stat.S_IRGRP)


def output_beacon(status):
    global logger

    done = False
    count = 0

    while not status.done:
        time.sleep(1)
        count += 1

        if count % BEACON_INTERVAL == 0:
            fmt = "Beacon: waited for another {sec} seconds..."
            logger.info(fmt.format(sec=BEACON_INTERVAL))


def execute_generated_script(p_main_setup_module, p_script_file_path_pattern):
    var = get_vars(p_setup_params=p_main_setup_module.extended_setup_params)

    output_file_path = p_script_file_path_pattern.format(**(var["setup"]))

    script_filename = os.path.join(get_module_dir(p_module=p_main_setup_module), output_file_path)

    timeout = p_main_setup_module.extended_setup_params.get("script_timeout")

    fmt = "<<<<< START script {filename} (timeout={timeout}[sec])..."
    logger.info(fmt.format(filename=script_filename, timeout=timeout))

    extended_env = os.environ.copy()
    extended_env.update(get_predefined_environment_variables())

    expand_vars(extended_env)
    thread = None
    status = None

    try:
        status = tools.SimpleStatus()
        thread = tools.start_simple_thread(output_beacon, status)

        process = subprocess.Popen(script_filename, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                   env=extended_env, bufsize=1, universal_newlines=True)

        msg = "[STDOUT/STDERR] {line}"

        for output in iter(process.stdout.readline, ""):
            for line in output.split("\n"):
                if line != '':
                    logger.info(msg.format(line=line))

        process.communicate(timeout=5)
        exit_code = process.returncode


    except subprocess.CalledProcessError as e:
        exit_code = e.returncode
        fmt = "Exception CalledProcessError in subprocess! stdout='{stdout}' stderr='{stderr}'"
        logger.error(fmt.format(stdout=e.output, stderr=e.stderr))

    except Exception as e:
        exit_code = -1
        msg = "General exception '{exception}' in subprocess!"
        logger.error(msg.format(exception=str(e)))

    finally:

        if status is not None:
            status.done = True

        if thread is not None:
            thread.join()


    msg = ">>>>> END script {filename} ..."
    logger.info(msg.format(filename=script_filename))

    if exit_code != 0:
        raise exceptions.ScriptExecutionError(p_script_name=script_filename, p_exit_code=exit_code)


def execute_make_debian_package_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=MAKE_DEBIAN_PACKAGE_SCRIPT_FILE_PATH)


def execute_build_docker_image_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=BUILD_DOCKER_IMAGE_SCRIPT_FILE_PATH)


def execute_install_debian_package_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=INSTALL_DEBIAN_PACKAGE_SCRIPT_FILE_PATH)


def execute_install_pypi_package_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=INSTALL_PYPI_PACKAGE_SCRIPT_FILE_PATH)


def execute_publish_debian_package_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=PUBLISH_DEBIAN_PACKAGE_SCRIPT_FILE_PATH)


def execute_publish_pypi_package_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=PUBLISH_PYPI_PACKAGE_SCRIPT_FILE_PATH)


def execute_test_app_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=TEST_APP_SCRIPT_FILE_PATH)

def execute_analyze_app_script(p_main_setup_module):
    execute_generated_script(p_main_setup_module=p_main_setup_module,
                             p_script_file_path_pattern=ANALYZE_APP_SCRIPT_FILE_PATH)

def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--execute-stage', dest='execute_stage', default=None,
                        help='execute CI stage',
                        choices=[STAGE_BUILD_PACKAGE,
                                 STAGE_BUILD_DOCKER_IMAGES,
                                 STAGE_PUBLISH_PACKAGE,
                                 STAGE_PUBLISH_PYPI_PACKAGE,
                                 STAGE_INSTALL,
                                 STAGE_INSTALL_PYPI_PACKAGE,
                                 STAGE_TEST,
                                 STAGE_ANALYZE,
                                 STAGE_TEARDOWN,
                                 STAGE_PREPARE])
    parser.add_argument('--run-dir', dest='run_dir', default=None)
    parser.add_argument('--use-dev-dir', dest='use_dev_dir', default=None)
    return parser


def main(p_main_module_dir):
    global logger

    exit_code = 0

    try:
        log_handling.start_logging()

        logger = log_handling.get_logger(__name__)

        parser = get_parser()

        arguments = parser.parse_args()

        main_setup_module = load_setup_module(p_dir=p_main_module_dir, p_module_name="setup")

        template_loader = jinja2.PackageLoader(python_base_app.__name__)
        template_env = jinja2.Environment(loader=template_loader)

        var = get_vars(p_setup_params=main_setup_module.extended_setup_params)

        if arguments.execute_stage == STAGE_BUILD_PACKAGE:
            if var["setup"]["build_debian_package"]:
                generate_debian_control(p_main_setup_module=main_setup_module, p_template_env=template_env)
                generate_debian_postinst(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                         p_arguments=arguments)
            generate_pip3_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                 p_arguments=arguments)
            generate_make_debian_package(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                         p_arguments=arguments)
            execute_make_debian_package_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_PUBLISH_PACKAGE:
            generate_publish_debian_package_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                                   p_arguments=arguments)
            execute_publish_debian_package_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_PUBLISH_PYPI_PACKAGE:
            generate_publish_pypi_package_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                                 p_arguments=arguments)
            execute_publish_pypi_package_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_BUILD_DOCKER_IMAGES:
            generate_build_docker_image_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                               p_arguments=arguments)
            execute_build_docker_image_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_INSTALL:
            generate_install_debian_package_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                                   p_arguments=arguments)
            execute_install_debian_package_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_INSTALL_PYPI_PACKAGE:

            generate_install_pypi_package_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                                 p_arguments=arguments)
            execute_install_pypi_package_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_TEST:
            generate_test_app_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                     p_arguments=arguments)
            generate_pycoveragerc(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                  p_arguments=arguments)
            execute_test_app_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_ANALYZE:
            generate_analyze_app_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                     p_arguments=arguments)
            execute_analyze_app_script(p_main_setup_module=main_setup_module)

        elif arguments.execute_stage == STAGE_PREPARE:
            generate_gitlab_ci_configuration(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                             p_arguments=arguments)
            generate_circle_ci_configuration(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                             p_arguments=arguments)
            generate_pip3_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                 p_arguments=arguments)
            generate_codacy_configuration(p_main_setup_module=main_setup_module, p_template_env=template_env)
            generate_git_python_file(p_main_setup_module=main_setup_module, p_template_env=template_env)

            var = get_vars(p_setup_params=main_setup_module.extended_setup_params)

            if var["setup"]["generate_generic_install"]:
                generate_generic_installation_script(p_main_setup_module=main_setup_module, p_template_env=template_env,
                                                     p_arguments=arguments)

        else:
            msg = "No stage selected -> nothing to do!"
            logger.warn(msg)

    except exceptions.ScriptExecutionError as e:
        fmt = "Execution of script {script_name} failed with exit code {exit_code}"
        logger.error(fmt.format(script_name=e.script_name, exit_code=e.exit_code))
        exit_code = 1

    except KeyError as e:
        fmt = "Exception of type {type}: {msg}"
        logger.exception(fmt.format(type=type(e), msg=str(e)))
        exit_code = 2

    except Exception as e:
        fmt = "General exception of type {type}: {msg}"
        logger.exception(fmt.format(type=type(e), msg=str(e)))
        exit_code = 2

    fmt = "Exiting with code {exit_code}"
    logger.info(fmt.format(exit_code=exit_code))

    return exit_code



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import inspect
import logging
import os
import sys
import unittest
from os.path import join, dirname

import pytest

from python_base_app import configuration
from python_base_app import log_handling

PARAMETER_NAME_CONFIGURATION_FILE = 'p_config_filename'
PARAMETER_NAME_TEST_DATA_BASE_DIR = 'p_test_data_base_dir'

DEFAULT_CONFIG_FILENAME = 'test.config'
DEFAULT_RESOURCE_REL_PATH = 'resources'

global g_capability_checkers

g_capability_checkers = {}


def register_capability_checkers(p_checkers):
    global g_capability_checkers

    g_capability_checkers.update(p_checkers)


class skip_if_env(object):

    def __init__(self, *p_env_names):

        if isinstance(p_env_names, list):
            self._env_names = p_env_names

        else:
            self._env_names = list(p_env_names)

    def __call__(self, f):

        def wrapped_f(*args):

            for env_name in self._env_names:
                if env_name in os.environ:
                    fmt = "Found {} in environment"
                    raise unittest.SkipTest(fmt.format(env_name))

            f(*args)

        return wrapped_f


class must_be_available(object):

    def __init__(self, p_required_capabilities):
        self._required_capabilities = p_required_capabilities
        self._logger = log_handling.get_logger(self.__class__.__name__)

    def __call__(self, f):

        def wrapped_f(*args):

            global g_capability_checkers

            missing_capabilities = []

            for cap in self._required_capabilities:
                if cap not in g_capability_checkers:
                    msg = "Capability '{cap}' has not been configured -> will be ignored"
                    self._logger.warn(msg.format(cap=cap))

                else:
                    checker = g_capability_checkers[cap]
                    if not checker():
                        missing_capabilities.append(cap)

            if len(missing_capabilities) > 0:
                fmt = "Missing capabilities: {caps}"
                msg = fmt.format(caps=", ".join(missing_capabilities))
                self._logger.info(msg)
                raise unittest.SkipTest(msg)

            f(*args)

        return wrapped_f


def add_tests_in_test_unit(
        p_test_suite, p_test_unit_class,
        p_config_filename=DEFAULT_CONFIG_FILENAME, p_test_data_base_dir=DEFAULT_RESOURCE_REL_PATH):
    for test_name in unittest.getTestCaseNames(p_test_unit_class, prefix="test"):
        test = p_test_unit_class(
            test_name,
            p_config_filename=p_config_filename,
            p_test_data_base_dir=p_test_data_base_dir)

        p_test_suite.addTest(test)


def get_config_filename():
    if len(sys.argv) > 1:
        return sys.argv[1]

    else:
        return None


def run_test_suite(p_test_suite):
    test_result = unittest.TextTestRunner(verbosity=2).run(p_test_suite)

    for (tc, reason) in test_result.skipped:
        print("SKIPPED: '%s'  REASON: '%s'" % (str(tc), reason))

    for (tc, reason) in test_result.errors:
        print("RUNTIME-ERROR: '%s'  REASON: '%s'" % (str(tc), reason))

    for (tc, reason) in test_result.failures:
        print("FAILED: '%s'  REASON: '%s'" % (str(tc), reason))

    if len(test_result.errors) > 0 or len(test_result.failures) > 0:
        exit(1)

    else:
        exit(0)


class BaseTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        base_dir = dirname(inspect.getfile(self.__class__))

        if PARAMETER_NAME_CONFIGURATION_FILE in kwargs:

            filename = kwargs.get(PARAMETER_NAME_CONFIGURATION_FILE, DEFAULT_CONFIG_FILENAME)

            if filename is None:
                filename = DEFAULT_CONFIG_FILENAME

            self._config_filename = join(base_dir, filename)
            del (kwargs[PARAMETER_NAME_CONFIGURATION_FILE])

        else:
            self._config_filename = join(base_dir, DEFAULT_CONFIG_FILENAME)

        if PARAMETER_NAME_TEST_DATA_BASE_DIR in kwargs:
            self._test_data_base_dir = join(base_dir, kwargs.get(PARAMETER_NAME_TEST_DATA_BASE_DIR))
            del (kwargs[PARAMETER_NAME_TEST_DATA_BASE_DIR])

        else:
            self._test_data_base_dir = join(base_dir, DEFAULT_RESOURCE_REL_PATH)

        super(BaseTestCase, self).__init__(*args, **kwargs)

        log_handling.start_logging(p_use_filter=False)
        self._config = self.configuration_factory()
        self._logger = log_handling.get_logger(self.__class__.__name__)

    @classmethod
    def setUpClass(cls):
        logger = log_handling.get_logger("urllib3.connectionpool")
        logger.setLevel(logging.WARNING)

    def configuration_factory(self):
        return configuration.Configuration()

    def add_config_section(self, p_section):

        self._config.add_section(p_section=p_section)

    def load_configuration(self):

        if self._config_filename is None:
            msg = "No filename specified for test case '{testcase}'"
            raise configuration.ConfigurationException(msg.format(testcase=self.__class__.__name__))

        self._config.read_config_file(self._config_filename,
                                      p_ignore_invalid_sections=True)

    def get_test_data_path(self, p_rel_path='.'):

        if self._test_data_base_dir is None:
            msg = "No base path specified for test case '{testcase}'"
            raise configuration.ConfigurationException(msg.format(testcase=self.__class__.__name__))

        return join(self._test_data_base_dir, p_rel_path)

    def execute_pytest(self, p_base_dir=None, p_rel_path="pytest"):

        if p_base_dir is None:
            p_base_dir = os.path.dirname(__file__)

        pytest_dir = os.path.join(p_base_dir, p_rel_path)

        fmt = "Executing pytest tests in directory {dirname}"
        self._logger.info(fmt.format(dirname=pytest_dir))

        pytest_result = pytest.cmdline.main([pytest_dir])

        self.assertEqual(pytest_result, 0)

    def check_list_length(self, p_list, p_length):

        self.assertIsNotNone(p_list)
        self.assertEqual(len(p_list), p_length)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import logging
#from os.path import join
from urllib.parse import urljoin

import requests

from python_base_app import configuration
from python_base_app import exceptions
from python_base_app import log_handling
from python_base_app import tools


class BaseRestAPIAccessConfigModel(configuration.ConfigModel):

    def __init__(self, p_section_name):
        super(BaseRestAPIAccessConfigModel, self).__init__(p_section_name)

        self.host_url = configuration.NONE_STRING

        self.login = configuration.NONE_STRING
        self.password = configuration.NONE_STRING
        self.access_token = configuration.NONE_STRING
        self.admin_login = configuration.NONE_STRING
        self.admin_password = configuration.NONE_STRING

    def is_active(self):
        return self.host_url is not None


class BaseRestAPIAccess(object):

    def __init__(self, p_config, p_section_name, p_base_api_url=""):

        self._config = p_config
        self._logger = log_handling.get_logger(p_section_name)

        # logger = log_handling.get_logger("requests.packages.urllib3.connectionpool")
        logger = log_handling.get_logger("urllib3.connectionpool")
        logger.setLevel(logging.WARNING)

        self._api_base_url = p_base_api_url

    def _get_api_url(self, p_command=None):

        if p_command is not None:
            return self._config.host_url + urljoin(self._api_base_url, p_command)

        else:
            return self._config.host_url + self._api_base_url

    def _handle_runtime_exception(self, p_status_code, p_exception, p_artifact_path='UNKNOWN', p_login='UNKNOWN',
                                  p_result_document=None, p_key=None):

        error_code = p_status_code
        error_message = None

        if p_artifact_path is not None:
            p_artifact_path = tools.anonymize_url(p_artifact_path)

        if error_code is None:
            try:
                error_json = json.loads(str(p_result_document))
                error_code = int(error_json['errors'][0]['status'])
                error_message = error_json['errors'][0]['message']

            except Exception:
                pass

        if error_code is None:
            try:
                error_json = json.loads(str(p_exception))
                error_code = int(error_json['errors'][0]['status'])
                error_message = error_json['errors'][0]['message']

            except Exception:
                pass

        if error_code is None:
            try:
                error_code = int(str(p_exception))

            except Exception:
                pass

        # print("error_message", error_message)

        if p_result_document is not None:
            fmt = "Result document: %s" % p_result_document
            self._logger.info(fmt)

        if error_code == 400 and error_message is not None and "key already exists" in error_message:

            fmt = "Objekt with key '%s' already exists" % p_key
            self._logger.error(fmt)
            raise exceptions.DuplicateKeyException(p_key=p_key, p_error_code=error_code)

        elif error_code == 401:

            if p_login is None:
                fmt = "Access to '{artifact_path}' was not authorized"

            else:
                fmt = "Access to '{artifact_path}' using login '{p_login}' was not authorized"

            self._logger.error(fmt.format(artifact_path=p_artifact_path, login=p_login))
            raise exceptions.UnauthorizedException(p_artifact_path=p_artifact_path, p_login=p_login,
                                                   p_error_code=error_code)

        elif error_code == 403:

            fmt = "Access to '%s' using login '%s' was not successful (too many attempts)" % (p_artifact_path, p_login)
            self._logger.error(fmt)
            raise exceptions.UnauthorizedException(p_artifact_path=p_artifact_path, p_login=p_login,
                                                   p_error_code=error_code)

        elif error_code == 404:

            fmt = "Artifact '%s' not found" % p_artifact_path
            self._logger.error(fmt)
            raise exceptions.ArtifactNotFoundException(p_artifact_path=p_artifact_path, p_error_code=error_code,
                                                       p_result_document=p_result_document)

        elif error_code == 416:

            fmt = "Range not satisfiable in '%s'" % p_artifact_path
            self._logger.error(fmt)
            raise exceptions.RangeNotSatisfiableException(p_artifact_path=p_artifact_path, p_error_code=error_code)

        elif error_code == 500 or error_code == 423:

            fmt = "Artifact '%s' is locked" % p_artifact_path
            self._logger.error(fmt)
            raise exceptions.ArtifactBlockedException(p_artifact_path=p_artifact_path, p_error_code=error_code)

        elif error_code == 504:

            fmt = "Timeout while accessing artifact '%s'" % p_artifact_path
            self._logger.error(fmt)
            raise exceptions.TimeoutException(p_artifact_path=p_artifact_path, p_error_code=error_code)

        else:

            fmt = "Unknown error code '%s' with message '%s' while accessing artifact '%s'" % (
                error_code, error_message, p_artifact_path)
            self._logger.error(fmt)

            if p_result_document is not None:
                fmt = "Result document: %s" % p_result_document
                self._logger.info(fmt)

            raise p_exception

    def _get_auth(self, p_requires_admin=False):
        if p_requires_admin:
            return tuple(self._config.admin_login, self._config.admin_password)

        else:
            return tuple(self._config.login, self._config.password)

    def execute_api_call(self, p_url, p_requires_authorization=False,
                         p_requires_admin=False,
                         p_mime_type=None,
                         p_method="GET",
                         p_data=None,
                         p_parameters=None,
                         p_jsonify=False):

        result = None
        headers = {}
        login = None
        r = None

        try:

            if p_requires_authorization:
                auth = self._get_auth(p_requires_admin)
                login = auth[0]

            else:
                auth = None

            if p_mime_type:
                headers['Content-type'] = p_mime_type

            fmt = "Executing %s API call '%s'" % (p_method, tools.anonymize_url(p_url))
            self._logger.debug(fmt)

            if p_method == "GET":
                r = requests.get(p_url, auth=auth, stream=False, headers=headers, params=p_parameters)

            elif p_method == "PUT":
                r = requests.put(p_url, auth=auth, data=p_data, headers=headers, params=p_parameters)

            elif p_method == "POST":
                r = requests.post(p_url, auth=auth, data=p_data, headers=headers, params=p_parameters)

            elif p_method == "DELETE":
                r = requests.delete(p_url, auth=auth, data=p_data, headers=headers, params=p_parameters)

            else:
                raise NotImplementedError("Invalid method '%s'" % p_method)

            try:
                if p_jsonify:
                    result = json.loads(r.content.decode("UTF-8"))

                else:
                    result = r.content.decode("UTF-8")
                # print ("result", result)

            except Exception:
                pass

            r.raise_for_status()

            r.close()


        except Exception as e:

            self._handle_runtime_exception(
                p_status_code=r.status_code if r is not None else None,
                p_exception=e,
                p_artifact_path=tools.anonymize_url(p_url),
                p_key=tools.anonymize_url(p_url),
                p_result_document=result,
                p_login=login)

        return result



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pwd

from python_base_app import base_user_handler
from python_base_app import configuration

SECTION_NAME = "UnixUserHandler"

HANDLER_NAME = "unix"


class UnixUserHandlerConfigModel(base_user_handler.BaseUserHandlerConfigModel):

    def __init__(self):
        super().__init__(p_section_name=SECTION_NAME)

        self.admin_username = configuration.NONE_STRING
        self.admin_password = configuration.NONE_STRING
        self.user_list = configuration.NONE_STRING

    def is_active(self):
        return self.admin_username is not None and self.admin_password is not None


class UnixUserHandler(base_user_handler.BaseUserHandler):

    def __init__(self, p_config, p_exclude_user_list=None):

        super().__init__(p_config=p_config, p_exclude_user_list=p_exclude_user_list)
        self._users = None

        msg = "Using admin user '{username}'"
        self._logger.info(msg.format(username=self._config.admin_username))

        if self._config.user_list is not None:
            try:
                self._users = {name_uid.split(":")[0].strip(): int(name_uid.split(":")[1]) for name_uid in
                               self._config.user_list.split(",")}

            except Exception as e:
                msg = "Invalid user:uid list '{list}' in [UnixUserHandler].user_list"
                raise configuration.ConfigurationException(msg.format(list=self._config.user_list))

            msg = "Using predefined users {users}"
            self._logger.info(msg.format(users=",".join(self._users.keys())))

    def list_users(self):

        users = []

        if self._users is None:
            for entry in pwd.getpwall():
                if self.is_valid_uid(p_uid=entry.pw_uid, p_username=entry.pw_name,
                                     p_password=entry.pw_passwd, p_shell=entry.pw_shell):
                    users.append(entry.pw_name)

            return users

        else:
            return self._users.keys()

    def get_uid(self, p_username):

        if self._users is None:
            try:
                user = pwd.getpwnam(p_username)

                if user is not None:
                    return user.pw_uid

            except KeyError:
                pass

        else:
            return self._users.get(p_username)

        return None

    def authenticate(self, p_username, p_password):

        return p_username == self._config.admin_username and p_password == self._config.admin_password

    def is_admin(self, p_username):

        return self._config.admin_username == p_username



# -*- coding: utf-8 -*-

# Copyright (C) 2019-2022  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

settings = {
    "name": "python-base-app",
    "version": "0.2.40",
    "description": "Generic tools and base classes for Python applications with web interface.",
    "author": "Marcus Rickert",
    "author_email": "marcus.rickert@web.de",
    "url": "https://github.com/marcus67/python_base_app",
}

extended_settings = {
    "babel_rel_directory": "translations"
}


# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import flask
import flask.wrappers
import flask_login

import python_base_app
import some_flask_helpers

AUTH_BLUEPRINT_NAME = 'auth'
AUTH_BLUEPRINT_ADAPTER = some_flask_helpers.BlueprintAdapter()

LOGIN_ENDPOINT_NAME = "login"
LOGOUT_ENDPOINT_NAME = "logout"

_ = lambda x: x

# for pybabel:
_('Please log in to access this page.')


class User(flask_login.UserMixin):

    def __init__(self, p_username):
        self.username = p_username
        self.id = p_username

    def get_id(self):
        return self.id


class AuthViewHandler(object):

    def __init__(self, p_user_handler, p_app, p_url_prefix, p_login_view=None,
                 p_logged_out_endpoint=LOGIN_ENDPOINT_NAME):

        if p_login_view is None:
            p_login_view = self.login_view()

        self._user_handler = p_user_handler
        self._logged_out_endpoint = p_logged_out_endpoint
        self._login_view = p_login_view
        self._url_prefix = p_url_prefix

        self._login_manager = flask_login.LoginManager()
        self._login_manager.init_app(app=p_app)
        self._login_manager.login_view = AUTH_BLUEPRINT_NAME + '.' + LOGIN_ENDPOINT_NAME
        self._login_manager.user_loader(self.load_user)

        # See https://flask-login.readthedocs.io/en/latest/#how-it-works
        self._login_manager.session_protection = "strong"

        self._blueprint = flask.Blueprint(AUTH_BLUEPRINT_NAME, python_base_app.__name__)
        AUTH_BLUEPRINT_ADAPTER.assign_view_handler_instance(p_blueprint=self._blueprint, p_view_handler_instance=self)
        AUTH_BLUEPRINT_ADAPTER.check_view_methods()

        p_app.register_blueprint(self._blueprint, url_prefix=p_url_prefix)

    def load_user(self, p_user_id):

        if self._user_handler.is_admin(p_user_id):
            return User(p_username=p_user_id)

        else:
            return None

    def check_user(self, p_username, p_password):

        if self._user_handler.authenticate(p_username=p_username, p_password=p_password):
            return User(p_username=p_username)

        else:
            return None

    def login_view(self):

        return flask.wrappers.Response('''
        <CENTER>
            <FORM ACTION="" METHOD="post">
                <TABLE>
                    <TR>
                        <TD ALIGN="RIGHT">Username:</TD>
                        <TD><INPUT TYPE=text NAME=username></TD>
                    </TR>
                    <TR>
                        <TD ALIGN="RIGHT">Password:</TD>
                        <TD><input type=password name=password></TD>
                    </TR>
                    <TR>
                        <TD COLSPAN="2" ALIGN="CENTER"><INPUT TYPE=submit VALUE=Login></TD>
                    </TR>
                </TABLE>
            </FORM>
        </CENTER>
        ''')

    @AUTH_BLUEPRINT_ADAPTER.route_method(p_rule='/login', endpoint=LOGIN_ENDPOINT_NAME, methods=["GET", "POST"])
    def login(self):

        request = flask.globals.request

        if flask.globals.request.method == 'POST':
            username = request.form['username']
            password = request.form['password']

            user = self.check_user(p_username=username, p_password=password)

            if user is not None:
                flask_login.login_user(user)
                next_arg = request.args.get("next")

                if next_arg is None:
                    return flask.redirect("")

                else:
                    # if next_arg.startswith(self._url_prefix):
                    #     next_arg = next_arg[len(self._url_prefix):]
                    return flask.redirect(next_arg)

            else:
                next_arg = request.args.get("next")
                flask.flash(_('Wrong login or password'))

                if next_arg is None:
                    return flask.redirect(flask.url_for("auth.login"))

                else:
                    return flask.redirect(flask.url_for("auth.login", next=next_arg))

        else:
            return self._login_view()

    @AUTH_BLUEPRINT_ADAPTER.route_method(p_rule='/logout', endpoint=LOGOUT_ENDPOINT_NAME, methods=["POST"])
    def logout(self):

        flask_login.logout_user()
        return flask.redirect(flask.url_for(self._logged_out_endpoint))

    def destroy(self):
        AUTH_BLUEPRINT_ADAPTER.unassign_view_handler_instances()



# -*- coding: utf-8 -*-

#    Copyright (C) 2021  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
import os
import re
import shlex
import subprocess

import requests

from python_base_app import configuration
from python_base_app import log_handling
from python_base_app import tools
from python_base_app.configuration import ConfigurationException

SECTION_NAME = "Pinger"

URL_SEPERATOR = ","

DEFAULT_PING_PORT = 6666

if tools.is_mac_os():
    DEFAULT_PING_COMMAND = "/sbin/ping"
    DEFAULT_PING_WAIT_OPTION = "-W"

else:
    DEFAULT_PING_COMMAND = "/bin/ping"
    DEFAULT_PING_WAIT_OPTION = "-w"

# MacOS: round-trip min/avg/max/stddev = 0.043/0.043/0.043/0.000 ms
# Linux: rtt min/avg/max/mdev = 0.043/0.043/0.043/0.000 ms

DEFAULT_PING_RESULT_REGEX = r"(rtt|round-trip) min/avg/max/(mdev|stddev) = [\d\.]+/([\d\.]+)/[\d\.]+/[\d\.]+ ms"
DEFAULT_PING_TIMEOUT_IN_SECONDS = 5


class PingerConfigModel(configuration.ConfigModel):

    def is_active(self):
        return True

    def __init__(self):
        super().__init__(p_section_name=SECTION_NAME)

        self.ping_command = DEFAULT_PING_COMMAND
        self.ping_result_regex = DEFAULT_PING_RESULT_REGEX
        self.ping_wait_option = DEFAULT_PING_WAIT_OPTION


class Pinger(object):

    def __init__(self, p_default_port=None, p_config=None):
        if p_config is None:
            p_config = PingerConfigModel()

        self._config = p_config
        self._default_port = p_default_port
        self._logger = log_handling.get_logger(self.__class__.__name__)

        try:
            self.ping_result_regex = re.compile(self._config.ping_result_regex)

        except Exception:
            fmt = "Invalid regular expression '{regex}' in [{section}]ping_result_regex"
            raise configuration.ConfigurationException(
                fmt.format(regex=self._config.ping_result_regex, section=SECTION_NAME))

    def is_valid_ping(self, p_host):

        if URL_SEPERATOR in p_host:
            first_url = p_host[:p_host.index(URL_SEPERATOR)]
            return tools.is_valid_dns_name(first_url)

        else:
            return tools.is_valid_dns_name(p_dns_name=p_host)

    def ping(self, p_host, p_default_port=None):

        if p_default_port is None:
            p_default_port = self._default_port

        if URL_SEPERATOR in p_host:
            return self.remote_ping(p_url=p_host, p_default_port=p_default_port)

        else:
            return self.local_ping(p_host=p_host)

    def remote_ping(self, p_url, p_default_port, p_default_timeout=DEFAULT_PING_TIMEOUT_IN_SECONDS):

        if not URL_SEPERATOR in p_url:
            msg = "No URL separator found in '{url}'!"
            raise Exception(msg.format(url=p_url))
        try:

            index = p_url.find(URL_SEPERATOR)
            first_url = p_url[:index]
            remaining_url = p_url[index + 1:]

            host, port = tools.split_host_url(p_url=first_url, p_default_port_number=p_default_port)

            url = "http://{host}:{port}/api/ping?host={remaining_url}".format(
                host=host,
                port=port,
                remaining_url=remaining_url
            )
            fmt = "Delegating to remote host using URL {url}"
            self._logger.debug(fmt.format(url=url))

            r = requests.get(url, timeout=p_default_timeout)

            fmt = "Result of remote ping: {delay}"
            self._logger.debug(fmt.format(delay=r.text))

            delay = float(r.text.replace(',', '.'))

        except Exception as e:
            fmt = "Exception during remote ping: {msg}"
            self._logger.error(fmt.format(msg=str(e)))
            return None

        return delay

    # https://stackoverflow.com/questions/2953462/pinging-servers-in-python
    def local_ping(self, p_host):
        """
        Checks if a host reacts to a ICMP ping request.
        Remember that a host may not respond to a ping (ICMP) request even if the host name is valid.

        :param p_host: DNS or ip address of the host to be pinged
        :return: the delay in milliseconds (as float) if the host responds, None otherwise
        """

        fmt = "{ping_command} {w_option} 1 {c_option} {host}"
        raw_command = fmt.format(ping_command=self._config.ping_command,
                                 # Ping command count option as function of OS
                                 c_option='-n 1' if tools.is_windows() else '-c 1',
                                 w_option=self._config.ping_wait_option,
                                 host=shlex.quote(p_host))

        command = shlex.split(raw_command)
        delay = None

        fmt = "Executing command {cmd} in Popen"
        self._logger.debug(fmt.format(cmd=command))

        # make sure we get floating points and not commas!
        run_env = os.environ.copy()
        run_env["LANG"] = "en_US"

        proc = subprocess.run(command, stdout=subprocess.PIPE, env=run_env)

        if proc.returncode >= 2:
            fmt = "{cmd} returns exit code {exitcode}"
            raise ConfigurationException(fmt.format(cmd=command, exitcode=proc.returncode))

        stdout_string = proc.stdout.decode("UTF-8")

        for line in stdout_string.split("\n"):
            fmt = "ping output: {line}"
            self._logger.debug(fmt.format(line=line))

            result = self.ping_result_regex.match(line)

            if result:
                delay = float(result.group(3))
                break

        fmt = "Host {host} is {status}"
        self._logger.debug(
            fmt.format(host=p_host, status="responding (%.1f ms)" % delay if delay is not None else "down"))

        return delay
